import airheadconfig

import email, email.Header, time, email.Utils, mailbox
from pysqlite2 import dbapi2 as sqlite # database
import os, stat

# our data source, brought to you by sqlite
connection = sqlite.connect(airheadconfig.dbname)
cursor = connection.cursor()
try:
	cursor.execute("select count(*) from devices")
except:	
	# whoops, probably a blank db - create tables

	# our internal device id <-> device data (ie, provided DeviceId)
	cursor.execute("""create table devices (
	  deviceid integer primary key autoincrement,
	  timeout integer,
	  devicedata text)""")
	cursor.execute("create index devicekey on devices (devicedata)")

	# mailbox name <-> mailbox id
	cursor.execute("""create table mailboxes (
	  mailboxid integer primary key autoincrement,
	  deviceid integer,
	  mailbox text,
	  synckey integer)""")
	cursor.execute("create index mailboxlookups on mailboxes (deviceid,mailbox)")
	
	# maildir: mail name <-> mail id
	cursor.execute("""create table mails (
	  deviceid integer,
	  mailboxid integer,
	  mailid text,
	  synckey integer,
	  remoteid integer primary key autoincrement,
	  read integer,
	  deleted integer,
	  time integer)""")
	# index for remote device lookups
	cursor.execute("create index remotelookups on mails (deviceid,mailboxid,remoteid)")
	# index for our own lookups
	cursor.execute("create index ownlookups on mails (deviceid,mailboxid,mailid)")
	
	connection.commit() # best to be safe ;p

def idForDeviceId(deviceid):
	# look for a pre-existing deviceid
	cursor.execute("select deviceid from devices where devicedata=?", (deviceid,))
	# if we found one, return it
	for x in cursor: return x[0]
	# otherwise, add a new one
	cursor.execute("insert into devices(devicedata) values(?)", (deviceid,))
	connection.commit()
	return idForDeviceId(deviceid) # muh, there must be a better way

def storeTimeout(deviceid, timeout):
	cursor.execute("update devices set timeout=? where deviceid=?", (timeout, deviceid))

def getTimeout(deviceid):
	cursor.execute("select timeout from devices where deviceid=?", (deviceid,))
	for i in cursor:
		return i[0]
	assert False, "couldn't find timeout for device " + str(deviceid) # shouldn't happen!

def detailsForMailbox(deviceid, mailboxname):
	# look for a pre-existing mailboxid
	mailboxquery = "select mailboxid,synckey from mailboxes where mailbox=? and deviceid=?"
	cursor.execute(mailboxquery, (mailboxname, deviceid))
	# if we found one, return it
	for x in cursor: return (x[0], x[1])

	# otherwise, add a new one
	cursor.execute("insert into mailboxes(synckey,deviceid,mailbox) values(?, ?, ?)", (0, deviceid, mailboxname))
	connection.commit()
	# try again
	cursor.execute(mailboxquery, (mailboxname, deviceid))
	for x in cursor: return (x[0], x[1])
	
	assert False, "failed to add db entry for " + mailboxname + " on device " + str(deviceid)

# todo: hacky, need a better solution!
def nameForMailCollection(deviceid, mailboxid):
  cursor.execute("select mailbox from mailboxes where deviceid=? and mailboxid=?", (deviceid, mailboxid))
  for x in cursor: return x[0]
  assert False, "failed to find db entry for mailbox #" + str(mailboxid) + " on device " + str(deviceid) # needs better error handling?

def setMailboxSyncKey(self, mboxid, synckey):
  cursor.execute("update mailboxes set synckey=? where mailboxid=?", (synckey, mboxid))
  connection.commit()

class MaildirSyncProvider:
  def __init__(self):
    self.collectioncache = {}

  def getCollection(self, deviceid, mboxid):
    if deviceid in self.collectioncache:
      if mboxid in self.collectioncache[deviceid]:
        return self.collectioncache[deviceid][mboxid]
    else:
      self.collectioncache[deviceid] = {}

    x = MaildirCollection(deviceid, mboxid)
    self.collectioncache[deviceid][mboxid] = x
    return x

  def sendMail(self, mailcontent):
    p = os.popen("/usr/sbin/sendmail -t -i", "w")
    p.write(mailcontent)
    p.close()

  def saveCopy(self, mailcontent, maildir):
    m = mailbox.Maildir(maildir)
    m.add(mailcontent)
    m.close()
    

  def getAttachment(self, deviceid, attachmentid):
    data = attachmentid.split(":")
    return self.getCollection(deviceid, data[0]).getAttachment(data[1])

# MaildirCollection represents the view of a device of a single maildir
class MaildirCollection:
  def __init__(self, deviceid, mboxid, mboxpath = None):
    self.deviceid = deviceid
    self.mboxid = mboxid
    if not mboxpath:
      mboxpath = nameForMailCollection(deviceid, mboxid)
    self.mboxpath = mboxpath

  def getChangedEmails(self, synckey, datefilter):
    emailfiles = os.listdir(self.mboxpath + "/new") + os.listdir(self.mboxpath + "/cur")
    emailids = map(lambda x: x.split(":")[0], emailfiles)
    tozap = [] # horror

    addedemails = []
    changedemails = []
    deletedemails = []

    if synckey != None:
      cursor.execute("select mailid,read,deleted,remoteid,time from mails where deviceid=? and mailboxid=? and synckey <= ?", (self.deviceid, self.mboxid, synckey))
    else:
      cursor.execute("select mailid,read,deleted,remoteid,time from mails where deviceid=? and mailboxid=?", (self.deviceid, self.mboxid))

    for i in cursor:
      mailid = i[0]
      read = i[1]
      deleted = i[2]
      remoteid = i[3]
      mailtime = i[4]
      if datefilter and mailtime and int(mailtime) < time.time() - datefilter: continue # too old
      if mailid in emailids:
        emailfile = emailfiles[emailids.index(mailid)]
        if emailfile in tozap: assert False, "ack, duplicate mail entries in the database" # sanity check
        if ":" in emailfile:
          flags = emailfile.split(":")[1]
        else:
          flags = ""

        if ("T" in flags) == (deleted == 0):
          # deleted status changed
          if not "T" in flags:
            continue # we should recreate this mail
          else:
            deletedemails.append(remoteid)
        elif ("S" in flags) == (read == 0):
          # read status changed
          changedemails.append((remoteid, ("S" in flags)))

        tozap.append(emailfile)
      elif not deleted:
        deletedemails.append(remoteid) # file is gone

    for i in tozap:
      emailfiles.remove(i)

    for i in emailfiles:
      # anything which remains is not in the database and therefore new
      mailid = i.split(":")[0]
      mailtime = os.stat(self.mboxpath + "/" + self.findFilenameFromMailname(mailid))[stat.ST_CTIME]
      if datefilter and mailtime < time.time() - datefilter: continue # too old
      if ":" in i:
        flags = i.split(":")[1]
      else:
        flags = ""
      if not "T" in flags:
        addedemails.append(i)

    return (addedemails, changedemails, deletedemails)

  def notifySentDelete(self, remoteid, synckey, deleted):
    # TODO: don't know how to handle this, if setting synckey is good, etc
    cursor.execute("update mails set deleted=?, synckey=? where remoteid=? and mailboxid=? and deviceid=?", (deleted, synckey, remoteid, self.mboxid, self.deviceid))
    connection.commit()
    
  def notifySentRead(self, remoteid, synckey, read):
    cursor.execute("update mails set read=?, synckey=? where remoteid=? and mailboxid=? and deviceid=?", (read, synckey, remoteid, self.mboxid, self.deviceid))
    connection.commit()

  def changeReadState(self, remoteid, synckey, read):
    self.changeEmailState(remoteid, synckey, "read", "S", read)
    self.notifySentRead(remoteid, synckey, read)
  
  def changeDeleteState(self, remoteid, synckey, delete):
    self.changeEmailState(remoteid, synckey, "deleted", "T", delete)
    self.notifySentDelete(remoteid, synckey, delete)

  def getInternalIdFromRemoteId(self, remoteid):
    emailfiles = os.listdir(self.mboxpath + "/new") + os.listdir(self.mboxpath + "/cur")
    cursor.execute("select mailid from mails where remoteid=? and mailboxid=? and deviceid=?", (remoteid, self.mboxid, self.deviceid))
    for i in cursor:
      for j in emailfiles:
        if j.split(":")[0] == i[0]:
          return j
    return None

  def findFilenameFromMailname(self, mailname):
    ourfile = None
    emailfiles = os.listdir(self.mboxpath + "/new") + os.listdir(self.mboxpath + "/cur")
    for i in emailfiles:
      if i.split(":")[0] == mailname:
        ourfile = i
    
    if not ourfile: return None

    # this is a bit horrible, we could do with something less horrid
    if ourfile in os.listdir(self.mboxpath + "/new"): ourfile = "new/" + ourfile
    else: ourfile = "cur/" + ourfile
    
    return ourfile

  def changeEmailState(self, remoteid, synckey, flagname, flagletter, newflag):
    cursor.execute("select mailid," + flagname + " from mails where remoteid=? and mailboxid=? and deviceid=?", (remoteid, self.mboxid, self.deviceid))
    data = cursor.fetchone()
    assert data # TODO: nicer error handling?

    if (newflag == data[1]): return # the database already matches? server version overrides, so we can always ignore this
    # TODO: maybe we should take notice of Conflicts and not always have server override..

    ourfile = self.findFilenameFromMailname(data[0])
    if not ourfile: return # file has disappeared

    # do the actual file move!
    newfile = ourfile

    if ":" in ourfile:
      basename = ourfile.split(":")[0]
      flags = ourfile.split(":")[1]
      if not flags.startswith("2,"):
        assert False, "help, we don't understand maildir flags " + flags
      flags = flags[2:]
      if "S" in flags and newflag == 0:
        flags = flags.replace(flagletter, "")
        if flags == "":
          newfile = basename
        else:
          newfile = basename + ":2," + flags
      elif (not flagletter in flags) and newflag == 1:
        # TODO: add in alphabetical order
        newfile = ourfile + flagletter
    else:
      if newflag:
        newfile = ourfile + ":2," + flagletter

    if newfile == ourfile: return

    # if we're touching a mail in new/, we need to move it into cur/
    newfilepath = newfile.split("/", 1)[0]
    if newfilepath == "new":
      newfile = "cur/" + newfile.split("/", 1)[1]

    # TODO: we should perhaps return if the move fails, to account for people reading read-only maildirs
    os.rename(self.mboxpath + "/" + ourfile, self.mboxpath + "/" + newfile)

  def markAsSent(self, mailid, synckey):
    mailid = mailid.split(":")[0]
    if ":" in mailid:
      read = "S" in mailid.split(":")[1]
    else:
      read = False
    t = os.stat(self.mboxpath + "/" + self.findFilenameFromMailname(mailid))[stat.ST_CTIME]

    cursor.execute("select remoteid from mails where mailid=? and deviceid=? and mailboxid=?", (mailid, self.deviceid, self.mboxid))
    for i in cursor:
      remoteid = i[0]
      # TODO: is this right?
      cursor.execute("update mails set synckey=?, deleted=? where remoteid=? and deviceid=? and mailboxid=?", (synckey, 0, remoteid, self.deviceid, self.mboxid))
      connection.commit()
      return remoteid # There is already a row in the database with this mailid.

    cursor.execute("insert into mails(mailid, deviceid, mailboxid, synckey, read, deleted, time) values(?, ?, ?, ?, ?, ?, ?)", (mailid, self.deviceid, self.mboxid, synckey, int(read), 0, t))
    connection.commit()

    cursor.execute("select remoteid from mails where deviceid=? and mailid=? and mailboxid=?", (self.deviceid, mailid, self.mboxid))
    for i in cursor:
      return i[0]
    assert False # wah!

  # internal function to find a good body from a message
  def findGoodBody(self, msg):
    if msg.is_multipart():
      gooddata = None
      for i in msg.get_payload():
        data = self.findGoodBody(i)
        if data[0] == "text/plain": return data
        elif data[0] == "text/html" and not gooddata: gooddata = data
      return gooddata
    else:
      return (msg.get_content_type(), msg.get_content_charset(), msg.get_payload(None, True))

  def findAttachments(self, msg):
    if msg.is_multipart():
      items = []
      for i in msg.get_payload():
        items = items + self.findAttachments(i)
    else:
      # TODO: get_filename doesn't seem reliable (doesn't catch Content-Type w/name, present in spam)
      # TODO: should we just ignore everything without a name?
      if msg.get_filename():
        payload = msg.get_payload(None, True)
        data = {'size': len(payload), 'type': msg.get_content_type(), 'name': msg.get_filename(), 'body': payload}
        items = [data]
      else:
        items = []

    return items

  # internal function to convert an email header to unicode
  def fixStupidEmailHeader(self, header):
    header = email.Header.decode_header(header)
    newheader = ""
    for i in header:
      # TODO: catch decode failures?
      if i[1] == None:
        newheader += i[0].decode("cp1252")
      else:
        newheader += i[0].decode(i[1])
    return newheader

  def getAttachment(self, attachmentid):
    # TODO: attachment names should be in a saner format?
    b = attachmentid.split("/")
    mailid = b[0]
    attid = int(b[1])

    ourfile = self.findFilenameFromMailname(mailid)
    if not ourfile: return None
    
    f = open(self.mboxpath + "/" + ourfile)
    m = email.message_from_file(f)
    att = self.findAttachments(m)
    assert attid < len(att), str(f) + " doesn't have a requested attachment#" + str(attid)
    attachment = att[attid]
    result = {'Body' : attachment['body']}
    if attachment['type']:
      result['Content-Type'] = attachment['type']
    return result

  # internal function to build an email
  def buildEmailInternal(self, mailid):
    read = False
    mailbits = mailid.split(":")
    if len(mailbits) > 1:
      read = "S" in mailbits[1]

    try:
      f = open(self.mboxpath + "/new/" + mailid)
    except:
      f = open(self.mboxpath + "/cur/" + mailid)
  
    m = email.message_from_file(f)

    body = self.findGoodBody(m)
    if body == None:
      bodytext = u"Couldn't find a body!"
      print bodytext
    else:
      bodytext = body[2]
      origcontenttype = body[1]
      contenttype = origcontenttype

      if not contenttype or contenttype in ["ascii", "us-ascii", "latin1", "iso8859-1"]:
        contenttype = "cp1252"
    
      try:
        bodytext = body[2].decode(contenttype)
      except Exception, e:
        bodytext = u"Failed to decode this body from " + origcontenttype + ": " + str(e)
        print bodytext

#    print f
    assert bodytext != None
    assert isinstance(bodytext, unicode)

    timestamp = "" # good default?
    if m["Date"]:
      parseddate = email.Utils.parsedate_tz(m["Date"])
      if parseddate:
        timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(email.Utils.mktime_tz(parseddate))) or ""

    # only DisplayTo *needs* to have the To header decoded, but best to decode all using fixStupidEmailHeader

    _to = self.fixStupidEmailHeader(m["To"] or "Unknown")
    _from = self.fixStupidEmailHeader(m["From"] or "Unknown")
    # TODO: cc, bcc?
    _subject = self.fixStupidEmailHeader(m["Subject"] or "")
    _read = 0
    if read: _read = 1

    values = { "To": _to,
      "From": _from,
      "Subject": _subject,
      "DateReceived": timestamp,
      "DisplayTo": _to,
      "ThreadTopic": _subject, # <-- TODO ("Thread-Topic" header?)
      "Read": str(_read),
      "Body": bodytext}

    if m["Reply-To"]: values["Reply-To"] = self.fixStupidEmailHeader(m["Reply-To"])
    if m["Cc"]: values["Cc"] = self.fixStupidEmailHeader(m["Cc"])

    att = self.findAttachments(m)
    if att:
      _id = 0
      for i in att:
        # TODO: attachment names should be in a saner format?
        i['id'] = str(self.mboxid) + ":" + mailbits[0] + "/" + str(_id)
        _id += 1
      values["Attachments"] = att
  
    return values

  # Return a dictionary representing an email, given a mailbox and a mailid.
  def buildEmail(self, mailid):
    d = self.buildEmailInternal(mailid)
    d["mailid"] = mailid

    if not "Importance" in d:
      d["Importance"] = "1"

    # fill in some dummy values

    # this is the message class; normally IPM.Note, delivery failures are REPORT.IPM.Note.NDR
    # http://msdn2.microsoft.com/en-us/library/aa125194.aspx has a list
    # TODO: should we try and detect things like delivery failures?
    d["MessageClass"] = "IPM.Note"

    # this is the Code Page of the email, it seems; 20127 for us-ascii
    # http://msdn2.microsoft.com/en-us/library/ms776446.aspx has a list
    # TODO: should we reply with the real codepage?
    d["InternetCPID"] = "65001" # utf-8

    return d

