#!/usr/bin/env python
# coding=utf-8

# airhead: an airsync push email server
# ♥ love from bertram and fuzzie! ♥
# passwords are sent in plaintext, so be sure to run this behind https!

from airheadconfig import configuration

from airheadbackend import * # TODO?
from airheadutils import *

from twisted.web2 import server, http, resource, channel, stream
from twisted.internet import reactor, defer
from twisted.python import log
from cStringIO import StringIO
from xml.dom import minidom
import pywbxml
import sys, os, base64

assert sys.version_info >= (2,5), "python 2.5 is required"

# log log log!
log.startLogging(sys.stdout)

syncprovider = MaildirSyncProvider()

# this is our one and only resource!
class Toplevel(resource.PostableResource):
  addSlash = True
  handlers = {}

  # all requests to the server should end up being handled by us
  def locateChild(self, request, segments):
    return (self, ())

  # verify a user is allowed access, based on a tuple of (domain, username, password)
  def verifyUser(self, userdetails):
    username = userdetails[1]
    if username in configuration:
      if 'password' in configuration[username]:
        if userdetails[2] == configuration[username]['password']:
          return True
    if airheadconfig.debug:
      print "auth failure: " + str(userdetails)
    return False

  # standard "go away" message for GET attempts
  def render(self, ctx):
    x = create_response(400)
    x.stream = stream.MemoryStream("Sorry, this server is unable to handle your request.")
    return x

  # handle OPTIONS requests
  def http_OPTIONS(self, req):
    resp = create_response()
    
    resp.headers.addRawHeader("Allow", "OPTIONS, POST")
    resp.headers.addRawHeader("Public", "OPTIONS, POST")
    resp.headers.addRawHeader("MS-ASProtocolVersions", "2.5")
    # for reference, Exchange 2007 sends "Sync,SendMail,SmartForward,SmartReply,GetAttachment,GetHierarchy,CreateCollection,DeleteCollection,MoveCollection,FolderSync,FolderCreate,FolderDelete,FolderUpdate,MoveItems,GetItemEstimate,MeetingResponse,Search,Settings,Ping,ItemOperations,Provision"
    resp.headers.addRawHeader("MS-ASProtocolCommands", "Sync,SendMail,SmartForward,SmartReply,GetAttachment,FolderSync,GetItemEstimate,Ping")
    return resp
 
  # initial handling of POST requests
  def http_POST(self, req):
    # TODO: perhaps this should be done in OPTIONS too?
    # if anyone enables anonymous access, you should set the username in authdata from the req
    authdata = decodeAuthorizationHeader(req.headers.getHeader('authorization'))
    if not (authdata and self.verifyUser(authdata)):
      return create_response(401)
      
    reply_deferred = defer.Deferred()
    buffer = StringIO()
    postdata = req.stream.read()
    if postdata:
    	postdata.addCallback(self.read_post_chunk, req, buffer, reply_deferred, authdata)
    else:
    	self.processPost(req, None, reply_deferred, authdata)
    return reply_deferred
  
  # callback for receiving POST data
  def read_post_chunk(self, chunk, req, buffer, reply_deferred, authdata):
    if chunk is not None:
      # add the data to the buffer, and add ourselves as the callback again
      buffer.write(chunk)
      postdata = req.stream.read()
      postdata.addCallback(self.read_post_chunk, req, buffer, reply_deferred, authdata)
    else:
      self.processPost(req, buffer.getvalue(), reply_deferred, authdata)

  # called when we've received all the POST data and are ready to do something  
  def processPost(self, req, postdata, reply_deferred, authdata):
      # we're done! parse the xbxml and then call handlePost for processing
      try:
        if not postdata or len(postdata) == 0:
          # no POST data
          data = None
        elif postdata[0] != chr(3): # first byte of wbxml is always 03
          # plain text
          data = postdata
        else:
          # wbxml
      	  data = minidom.parseString(pywbxml.wbxml2xml(postdata))
      	  if airheadconfig.debug:
            print data.toprettyxml()

        reply = self.handlePost(req, data, reply_deferred, authdata)
      except:
        reply = create_response(500) # internal server error
        sys.excepthook(*sys.exc_info()) # display traceback on stderr
      
      if reply != reply_deferred:
        reply_deferred.callback(reply)

  # do the actual handling of a POST request
  def handlePost(self, req, data, reply_deferred, authdata):
    cmd = req.args["Cmd"][0]
    devicedata = req.args["DeviceId"][0]
    deviceid = idForDeviceId(devicedata)
    
    if airheadconfig.debug:
      print "trying to handle '" + cmd + "' for user '" + authdata[1] + "' using device id: " + str(deviceid) + " (" + devicedata + ")"
    
    if cmd == "FolderSync":
      return self.handleFolderSync(deviceid, data, authdata)
    elif cmd == "GetItemEstimate":
      return self.handleGetItemEstimate(deviceid, data)
    elif cmd == "Sync":
      return self.handleSync(deviceid, data)
    elif cmd == "Ping":
      return self.handlePing(deviceid, data, reply_deferred, authdata)
    elif cmd == "SendMail":
      return self.handleSendMail(data, req, authdata)
    elif cmd == "SmartReply":
      return self.handleSmartReply(deviceid, data, req, authdata)
    elif cmd == "SmartForward":
      return self.handleSmartForward(deviceid, data, req, authdata)
    elif cmd == "GetAttachment":
      return self.handleGetAttachment(deviceid, data, req)
    else:
      log.msg("warning: failed to understand command " + cmd + ", erroring")
      return create_response(400) # bad request

  def handleGetAttachment(self, deviceid, data, req):
    attachmentname = req.args["AttachmentName"][0]
    attachmentname = base64.b64decode(attachmentname)
    attachment = syncprovider.getAttachment(deviceid, attachmentname)
    if not attachment:
      return create_response(400) # TODO: better error?

    resp = create_response()
    if 'Content-Type' in attachment:
      resp.headers.addRawHeader("Content-Type", attachment['Content-Type'])
    if 'Stream' in attachment:
      resp.stream = attachment['Stream']
    else:
      assert 'Body' in attachment
      resp.stream = stream.MemoryStream(attachment['Body'])
    return resp

  # convenience function for mangleMail
  def calculateFromAddress(self, authdata):
    if 'defaultfrom' in configuration[authdata[1]]:
      fromaddress = configuration[authdata[1]]['defaultfrom']
    else:
      fromaddress = authdata[1]
      if authdata[0]: fromaddress = fromaddress + "@" + authdata[0]
    return fromaddress

  # helper function for SendMail etc
  def mangleMail(self, maildata, authdata):
    message = email.message_from_string(maildata.encode("utf-8"))
    message.add_header("X-Mailer", "airsync client via airhead sync server")
    fromaddress = self.calculateFromAddress(authdata)
    if "From" in message.keys():
      message.replace_header("From", fromaddress)
    else:
      message.add_header("From", fromaddress)

    return message

  def handleSendMail(self, maildata, req, authdata):
    assert maildata != None

    msg = self.mangleMail(maildata, authdata).as_string(False)

    syncprovider.sendMail(msg)

    if req.args["SaveInSent"][0]:
      if configuration[authdata[1]].has_key('sent_mail'):
        syncprovider.saveCopy(msg, configuration[authdata[1]]['sent_mail'])
      elif airheadconfig.debug:
        print "Should be saving the sent email, but sent_mail is unset"

    return create_response(200)

  def handleSmartReply(self, deviceid, maildata, req, authdata):
    collection = syncprovider.getCollection(deviceid, req.args["CollectionId"][0])
    mail = collection.buildEmail(collection.getInternalIdFromRemoteId(req.args["ItemId"][0].split(":")[1]))
    body = maildata.strip() + "\n\n"
    body += "\n".join(map(lambda x: "> " + x, mail['Body'].strip().split("\n")))
    return self.handleSendMail(body, req, authdata);

  def handleSmartForward(self, deviceid, maildata, req, authdata):
    collection = syncprovider.getCollection(deviceid, req.args["CollectionId"][0])
    mail = collection.buildEmail(collection.getInternalIdFromRemoteId(req.args["ItemId"][0].split(":")[1]))
    body = maildata.strip() + "\n"
# For some reason the forward doesn't include the attribution
    body += "\n---- Forwarded message: ----\n"
    body += "From: " + mail["From"] + "\n"
    body += "To: " + mail["DisplayTo"] + "\n"
    body += "Date: " + mail["DateReceived"] + "\n"
    body += "Subject: " + mail["Subject"] + "\n\n"
    body += "\n".join(map(lambda x: "> " + x, mail['Body'].strip().split("\n")))
    return self.handleSendMail(body, req, authdata);

  def handleGetItemEstimate(self, deviceid, xmldata):
    assert xmldata != None
    # parse incoming xml for collections containing a collection id, a synckey and a class
    syncrequests = []
    collections = xmldata.getElementsByTagName("Collection")
    for i in collections:
      collectionid = i.getElementsByTagName("CollectionId")
      # FIXME Duplicated code
      filtertype = i.getElementsByTagName("FilterType")
      if filtertype:
        datefilter = (0, 1, 3, 7, 14, 31, 31, 31, 31)[int(filtertype[0].firstChild.nodeValue)] * 60 * 60 * 24
      if airheadconfig.debug: print "Filter value " + filtertype[0].firstChild.nodeValue + ", only showing last " + str(datefilter) + " seconds"
      if collectionid: collectionid = int(collectionid[0].firstChild.nodeValue)
      synckey = int(i.getElementsByTagName("SyncKey")[0].firstChild.nodeValue)
      classname = i.getElementsByTagName("Class")[0].firstChild.nodeValue
      syncrequests.append({'collectionid': collectionid, 'synckey': synckey, 'classname': classname, 'datefilter': datefilter})
                                            
    # calculate the item estimate
    for i in syncrequests:
      assert i['classname'] == "Email" or not i['collectionid']

      if i['collectionid']:
        mailbox = syncprovider.getCollection(deviceid, i['collectionid'])
        changedemails = mailbox.getChangedEmails(i['synckey'], i['datefilter'])
        i['estimate'] = len(changedemails[0]) + len(changedemails[1]) + len(changedemails[2])
        if airheadconfig.debug:
          print "going to claim that " + str(i['estimate']) + " item" + "s" * (i['estimate'] != 1) + " need to be synced for " + mailbox.mboxpath

    # create the result xml
    d = create_airsync_document("GetItemEstimate")

    for i in syncrequests:
      responseNode = d.createElement("Response")
      d.documentElement.appendChild(responseNode)
      statusNode = d.createElement("Status")
      statusNode.appendChild(d.createTextNode("1"))
      responseNode.appendChild(statusNode)

      collectionNode = d.createElement("Collection")
      responseNode.appendChild(collectionNode)

      if i['collectionid']:
        collectionIdNode = d.createElement("CollectionId")
        collectionIdNode.appendChild(d.createTextNode(str(i['collectionid'])))
        collectionNode.appendChild(collectionIdNode)

      classNode = d.createElement("Class")
      classNode.appendChild(d.createTextNode(i['classname']))
      collectionNode.appendChild(classNode)

      estimateNode = d.createElement("Estimate")
      if i['collectionid']: # hack!
        estimateNode.appendChild(d.createTextNode(str(i['estimate'])))
      else:
        estimateNode.appendChild(d.createTextNode(str(0)))
      collectionNode.appendChild(estimateNode)

#   print d.toprettyxml()
    return create_airsync_response(d)

  def handleFolderSync(self, deviceid, xmldata, authdata):
    assert xmldata != None
    synckey = int(xmldata.getElementsByTagName("SyncKey")[0].firstChild.nodeValue)

    addedmailboxes = []
    deledmailboxes = []

    # TODO: check for mailbox changes since last synckey
    if synckey == 0:
      assert 'datasources' in configuration[authdata[1]], "configuration has no data sources for user " + authdata[1]
      datasources = configuration[authdata[1]]['datasources']
      for i in datasources.keys():
        mailboxinfo = detailsForMailbox(deviceid, datasources[i])
        if i == 'inbox':
          mailboxtype = 2
        else:
          mailboxtype = 12
        addedmailboxes.append([mailboxinfo[0], i, mailboxtype])
      synckey += 1

    # TODO: call setMailboxSyncKey as necessary

    d = create_airsync_document("FolderSync")
    statusNode = d.createElement("Status")
    statusNode.appendChild(d.createTextNode("1"))
    d.documentElement.appendChild(statusNode)
    syncKeyNode = d.createElement("SyncKey")
    syncKeyNode.appendChild(d.createTextNode(str(synckey)))
    d.documentElement.appendChild(syncKeyNode)
    changesNode = d.createElement("Changes")
    countNode = d.createElement("Count")
    countNode.appendChild(d.createTextNode(str(len(addedmailboxes) + len(deledmailboxes))))
    changesNode.appendChild(countNode)
    for i in addedmailboxes:
      a = d.createElement("Add")
      s = d.createElement("ServerId")
      s.appendChild(d.createTextNode(str(i[0])))
      p = d.createElement("ParentId")
      p.appendChild(d.createTextNode("0"))
      dn = d.createElement("DisplayName")
      dn.appendChild(d.createTextNode(i[1]))
      t = d.createElement("Type")
      t.appendChild(d.createTextNode(str(i[2])))
      changesNode.appendChild(a)
      for j in (s, p, dn, t):
        a.appendChild(j)    
    # TODO: deledmailboxes
    d.documentElement.appendChild(changesNode)
#   print d.toprettyxml()
    return create_airsync_response(d)

  def handlePing(self, deviceid, xmldata, reply_deferred, authdata):
    if xmldata and xmldata.getElementsByTagName("HeartbeatInterval"):
      timeout = int(xmldata.getElementsByTagName("HeartbeatInterval")[0].firstChild.nodeValue)
      storeTimeout(deviceid, timeout)
    else:
      timeout = getTimeout(deviceid)
      if timeout == None:
        print "warning: no heartbeat interval in the database for this device, defaulting to 200s"
        timeout = 200; # who knows?

    if airheadconfig.debug:
      print "starting ping, timeout " + str(timeout) + "s"
    
    # cancel any existing ping bits
    # TODO: ideally we should do this on connection drop, too, but fuzzie can't work out how to detect such an event
    try:
      if deviceid in self.handlers:
        self.handlers[deviceid][0].cancel()
        self.handlers[deviceid][1].cancel()
    except:
      pass
    
    # store the handlers for later cancellation/checks
    # we assume a single deviceid is only doing one Ping at a time, which i think is reasonable
    self.handlers[deviceid] = []
    self.handlers[deviceid].append(reactor.callLater(timeout, self.handlePingReply, deviceid, reply_deferred, []))
    self.handlers[deviceid].append(reactor.callLater(5, self.handlePingCheck, deviceid, reply_deferred, authdata, self.handlers[deviceid][0]))
    
    return reply_deferred

  # poll to see if we need to reply to Ping yet
  def handlePingCheck(self, deviceid, reply_deferred, authdata, timeouthandler):
    if timeouthandler.called: return # timeout has happened
    if airheadconfig.debug:
      print "doing ping check.."
    assert 'datasources' in configuration[authdata[1]], "configuration has no data sources for user " + authdata[1]
    datasources = configuration[authdata[1]]['datasources']
    # TODO: hardcoding inbox is badness
    details = detailsForMailbox(deviceid, datasources['inbox'])
    collection = syncprovider.getCollection(deviceid, details[0])
    emailids = collection.getChangedEmails(None, None)
    if len(emailids[0]): # new emails
      if airheadconfig.debug:
        print "new email during ping! replying.."
      timeouthandler.cancel() # no need for this any more
      self.handlePingReply(deviceid, reply_deferred, [details[0]])
    else:
      self.handlers[deviceid][1] = reactor.callLater(5, self.handlePingCheck, deviceid, reply_deferred, authdata, timeouthandler)

  # handle the final Ping reply by throwing out some xml
  def handlePingReply(self, deviceid, reply_deferred, mailboxes):
    if mailboxes: status = 2
    else: status = 1
    d = create_airsync_document("Ping")
    s = d.createElement("PingStatus") # TODO: work out how to fix wbxml2
    s.appendChild(d.createTextNode(str(status)))
    d.documentElement.appendChild(s)
    for i in mailboxes:
      f = d.createElement("PingFolders")
      fi = d.createElement("PingFolderId")
      d.documentElement.appendChild(f)
      f.appendChild(fi)
      fi.appendChild(d.createTextNode(str(i)))
    if airheadconfig.debug:
      print d.toprettyxml()
    reply_deferred.callback(create_airsync_response(d))

  # create an email node suitable for use when Syncing with Add or Fetch commands
  def makeEmailNode(self, d, nodename, collectionid, remoteid, data, maxsize):
    main = d.createElement(nodename)

    serverid = d.createElement("ServerId")
    serverid.appendChild(d.createTextNode(str(collectionid) + ":" + str(remoteid)))
    main.appendChild(serverid)

    if nodename == "Fetch":
      status = d.createElement("Status")
      status.appendChild(d.createTextNode("1"))
      main.appendChild(status)

    data["BodyTruncated"] = "0"

    if maxsize != None:
      # TODO: these are unicode strings, so we're work with characters here, not bytes, which is likely wrong
      if len(data["Body"]) > maxsize:
        data["BodySize"] = str(len(data["Body"]))
        data["Body"] = data["Body"][:maxsize]
        data["BodyTruncated"] = "1"

    appdata = d.createElement("ApplicationData")
    main.appendChild(appdata)
    for j in ["To", "Cc", "From", "Subject", "Reply-To", "DateReceived", "DisplayTo", "ThreadTopic", "Importance", "Read", "BodyTruncated", "BodySize", "Body", "Attachments", "MessageClass", "InternetCPID"]:
      if j in data:
        if j == "Attachments":
          appdata.appendChild(self.makeAttachmentsNode(d, data[j]))
        else:
          element = d.createElement(j)
          element.appendChild(d.createTextNode(data[j]))
          appdata.appendChild(element)
    return main
  
  def makeAttachmentsNode(self, d, data):
    attachments = d.createElement("Attachments")
    for i in data:
      att = d.createElement("Attachment")
      attachments.appendChild(att)

      # we're doing ATTACH_BY_VALUE (1)
      # TODO: do we ever want to do ATTACH_EMBEDDED_MSG (5)?
      attmethod = d.createElement("AttMethod")
      attmethod.appendChild(d.createTextNode("1"))
      att.appendChild(attmethod)
      
      attsize = d.createElement("AttSize")
      attsize.appendChild(d.createTextNode(str(i['size'])))
      att.appendChild(attsize)
      
      displayname = d.createElement("DisplayName")
      displayname.appendChild(d.createTextNode(i['name']))
      att.appendChild(displayname)
      
      attname = d.createElement("AttName")
      attname.appendChild(d.createTextNode(base64.b64encode(i['id'])))
      att.appendChild(attname)
      
    return attachments

  def handleSync(self, deviceid, xmldata):
    assert xmldata != None
        
    d = create_airsync_document("Sync")
    c = d.createElement("Collections")
    d.documentElement.appendChild(c)
    collections = xmldata.getElementsByTagName("Collection")
    for i in collections:
      collectionid = i.getElementsByTagName("CollectionId")
      if collectionid: collectionid = int(collectionid[0].firstChild.nodeValue)
      synckey = int(i.getElementsByTagName("SyncKey")[0].firstChild.nodeValue)

      windowsize = i.getElementsByTagName("WindowSize")
      if windowsize:
        windowsize = int(windowsize[0].firstChild.nodeValue)

      getchanges = i.getElementsByTagName("GetChanges")

      # default options
      datefilter = None
      maxsize = None
      serveroverrules = True

      # read options from xml
      options = i.getElementsByTagName("Options")
      if options:
        for opt in options:
          filtertype = opt.getElementsByTagName("FilterType")
          if filtertype:
            # None, 1 day, 3 days, a week, 2 weeks, a month, ??, ??, ??
            datefilter = (0, 1, 3, 7, 14, 31, 31, 31, 31)[int(filtertype[0].firstChild.nodeValue)] * 60 * 60 * 24
          truncation = opt.getElementsByTagName("Truncation")
          if truncation:
            # we don't know the values for 3 or 5-8, so we set them to the same as 4 (ie, 5kb)
            truncationsettings = [0, 500, 1000, 5000, 5000, 5000, 5000, 5000, 5000, None]
            maxsize = truncationsettings[int(truncation[0].firstChild.nodeValue)]
          conflict = opt.getElementsByTagName("Conflict")
          if conflict:
            serveroverrules = (conflict[0].firstChild.nodeValue == "1")

      classname = i.getElementsByTagName("Class")[0].firstChild.nodeValue
      assert classname == "Email" or not collectionid

      if collectionid and getchanges:
        collection = syncprovider.getCollection(deviceid, collectionid)

      tofetch = []

      commands = i.getElementsByTagName("Commands")
      if commands and collectionid: # TODO: should handle things except email
        synckey += 1
        
        # TODO: handle Add?
        
        fetches = commands[0].getElementsByTagName("Fetch")
        for i in fetches:
          serverid = i.getElementsByTagName("ServerId")[0].firstChild.nodeValue
          serverid = serverid.split(":")
          assert serverid[0] == str(collectionid)
          serverid = int(serverid[1])
          
          tofetch.append(serverid)

        deletions = commands[0].getElementsByTagName("Delete")
        for i in deletions:
          serverid = i.getElementsByTagName("ServerId")[0].firstChild.nodeValue
          serverid = serverid.split(":")
          assert serverid[0] == str(collectionid)
          serverid = serverid[1]

          collection.changeDeleteState(serverid, synckey, 1)

        changes = commands[0].getElementsByTagName("Change")
        for i in changes:
          serverid = i.getElementsByTagName("ServerId")[0].firstChild.nodeValue
          serverid = serverid.split(":")
          assert serverid[0] == str(collectionid)
          serverid = serverid[1]

          appdata = i.getElementsByTagName("ApplicationData")[0]

          # we assume email from here on :)
          read = appdata.getElementsByTagName("Read")
          if not read:
            # Read is the only valid change for emails in protocol 2.5
            assert False, "ack, got a client Change without a Read in it"
          read = int(read[0].firstChild.nodeValue)
          collection.changeReadState(serverid, synckey, read)

      if collectionid and getchanges:
        changedemails = collection.getChangedEmails(synckey, datefilter)
        numchangedemails = len(changedemails[0]) + len(changedemails[1]) + len(changedemails[2])

      # don't increment the key if we've already incremented it before commands..
      if (collectionid and getchanges and numchangedemails) and (not commands):
        synckey += 1

      co = d.createElement("Collection")
      cl = d.createElement("Class")
      cl.appendChild(d.createTextNode(classname))
      sy = d.createElement("SyncKey")
      sy.appendChild(d.createTextNode(str(synckey)))
      if collectionid:
        ci = d.createElement("CollectionId")
        ci.appendChild(d.createTextNode(str(collectionid)))
      else:
        ci = None
      st = d.createElement("Status")
      st.appendChild(d.createTextNode("1"))
      for j in (cl, sy, ci, st):
        if j: co.appendChild(j)
      c.appendChild(co)

      if not collectionid: continue

      if getchanges and numchangedemails:
        # TODO: we only apply windowsize to added emails here, should we apply to other things?
        if windowsize and (len(changedemails[0]) > windowsize):
          if airheadconfig.debug:
            print "more than window size of " + str(windowsize) + " mails available; sending MoreAvailable"
          more = d.createElement("MoreAvailable")
          co.appendChild(more)

        cm = d.createElement("Commands")
        co.appendChild(cm)

        # construct Change objects
        for i in changedemails[1]:
          remoteid = i[0]
          read = i[1]

          change = d.createElement("Change")
          cm.appendChild(change)

          serverid = d.createElement("ServerId")
          serverid.appendChild(d.createTextNode(str(collectionid) + ":" + str(remoteid)))
          change.appendChild(serverid)

          appdata = d.createElement("ApplicationData")
          readnode = d.createElement("Read")
          readnode.appendChild(d.createTextNode(str(int(read))))
          appdata.appendChild(readnode)
          change.appendChild(appdata)

          collection.notifySentRead(remoteid, synckey, int(read))

        # construct Delete objects
        for i in changedemails[2]:
          delete = d.createElement("Delete")
          cm.appendChild(delete)

          serverid = d.createElement("ServerId")
          serverid.appendChild(d.createTextNode(str(collectionid) + ":" + str(i)))
          delete.appendChild(serverid)
            
          collection.notifySentDelete(i, synckey, 1)

        # construct Add objects
        addeditems = changedemails[0]
        if windowsize: addeditems = addeditems[:windowsize]
        for i in addeditems:
          i = collection.buildEmail(i)
          remoteid = collection.markAsSent(i["mailid"], synckey)
          print "sending " + i["mailid"] + " as " + str(remoteid)
          cm.appendChild(self.makeEmailNode(d, "Add", collectionid, remoteid, i, maxsize))

      if len(tofetch):
        cm = d.createElement("Responses")
        co.appendChild(cm)
        
        # construct Fetch objects
        for i in tofetch:
          # TODO: getInternalIdFromRemoteId is icky, the backend should hide itself better, perhaps
          remoteid = i
          internalid = collection.getInternalIdFromRemoteId(remoteid)
          # TODO: print warning otherwise?
          if internalid:
            i = collection.buildEmail(internalid)
            assert remoteid == collection.markAsSent(i["mailid"], synckey)
            if airheadconfig.debug:
              print "sending fetched " + i["mailid"] + " as " + str(remoteid)
            cm.appendChild(self.makeEmailNode(d, "Fetch", collectionid, remoteid, i, None))

#    print d.toxml().encode('utf-8')
    return create_airsync_response(d)

# start twisted server gubbins
t = Toplevel()
site = server.Site(t)
factory = channel.HTTPFactory(site)
try:
  reactor.listenTCP(80, factory)
except:
  reactor.listenTCP(8080, factory)
reactor.run()
