#!/usr/bin/env python
# coding=utf-8

import base64
from twisted.web2 import http, stream
from xml.dom import minidom
import pywbxml

# convenience function for creating an airsync document via minidom
def create_airsync_document(rootnode):
  doctype = minidom.getDOMImplementation().createDocumentType("AirSync", "-//AIRSYNC//DTD AirSync//EN", "http://www.microsoft.com/")
  return minidom.getDOMImplementation().createDocument(None, rootnode, doctype)

# convenience function to extract a tuple of (domain, username, password) from a basic auth header
def decodeAuthorizationHeader(authdata):
  if authdata == None or len(authdata) != 2 or authdata[0] != "basic":
    return None
  else:
    passdata = base64.b64decode(authdata[1]).split(":", 1)
    if len(passdata) != 2:
      return None
    else:
      userdata = passdata[0].split("\\", 1)
      if len(userdata) > 1:
        domain = userdata[0]
      else:
        domain = None
      username = userdata[-1]
      password = passdata[1]
      return (domain, username, password)

# convenience function to create an http.Response object
def create_response(code = 200):
  resp = http.Response(code)
  resp.headers.addRawHeader("MS-Server-ActiveSync", "8.0") # TODO: this seems to come out in lowercase .. ?
  resp.headers.addRawHeader("Server", "airhead/0.1.0") # TODO: suck version number from somewhere
  resp.stream = stream.MemoryStream("") # empty by default
  return resp
  
# convenience function to create an airsync response from a minidom document
def create_airsync_response(doc):
  resp = create_response()
  resp.headers.addRawHeader("Content-Type", "application/vnd.ms-sync.wbxml")
  resp.stream = stream.MemoryStream(pywbxml.xml2wbxml(doc.toxml()))
  #print "sent xml:"
  #print pywbxml.xml2wbxml(doc.toxml())
  #print minidom.parseString(pywbxml.wbxml2xml(pywbxml.xml2wbxml(doc.toxml()))).toprettyxml()
  return resp
